@@ -1,7 +1,6 @@
 # MIT licensed
 # Copyright (c) 2013-2020 lilydjwg <lilydjwg@gmail.com>, et al.
 
-import atexit
 import asyncio
 from typing import Optional, Dict
 
@@ -16,16 +15,15 @@ logger = structlog.get_logger(logger_name=__name__)
 connector = aiohttp.TCPConnector(limit=20)
 
 class AiohttpSession(BaseSession):
+  session = None
+
   def setup(
     self,
     concurreny: int = 20,
     timeout: int = 20,
   ) -> None:
-    self.session = aiohttp.ClientSession(
-      connector = aiohttp.TCPConnector(limit=concurreny),
-      timeout = aiohttp.ClientTimeout(total=timeout),
-      trust_env = True,
-    )
+    self._concurreny = concurreny
+    self._timeout = timeout
 
   async def request_impl(
     self, url: str, *,
@@ -35,8 +33,17 @@ class AiohttpSession(BaseSession):
     follow_redirects: bool = True,
     params = (),
     json = None,
+    body = None,
     verify_cert: bool = True,
   ) -> Response:
+    if self.session is None:
+      # need to create in async context
+      self.session = aiohttp.ClientSession(
+        connector = aiohttp.TCPConnector(limit=self._concurreny),
+        timeout = aiohttp.ClientTimeout(total=self._timeout),
+        trust_env = True,
+      )
+
     kwargs = {
       'headers': headers,
       'params': params,
@@ -47,7 +54,13 @@ class AiohttpSession(BaseSession):
 
     if proxy is not None:
       kwargs['proxy'] = proxy
-    if json is not None:
+    if body is not None:
+      # Make sure all backends have the same default encoding for post data.
+      if 'Content-Type' not in headers:
+        headers = {**headers, 'Content-Type': 'application/x-www-form-urlencoded'}
+        kwargs['headers'] = headers
+      kwargs['data'] = body.encode()
+    elif json is not None:
       kwargs['json'] = json
 
     try:
@@ -70,9 +83,4 @@ class AiohttpSession(BaseSession):
     body = await res.content.read()
     return Response(res.headers, body)
 
-@atexit.register
-def cleanup():
-  loop = asyncio.get_event_loop()
-  loop.run_until_complete(session.session.close())
-
 session = AiohttpSession()
